@@ -6,19 +6,8 @@ Authors of original libkmahjongg in C++:
 this python code:
     Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
-    Kajongg is free software you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation either version 2 of the License, or
-    (at your option) any later version.
+SPDX-License-Identifier: GPL-2.0
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
 from qt import Qt, QPainter, QBrush, QPalette, QPixmapCache, QPixmap
@@ -38,7 +27,7 @@ class Background(Resource):
 
     def __init__(self, name):
         """continue __build"""
-        super(Background, self).__init__(name)
+        super().__init__(name)
         self.__svg = None
         self.__pmap = None
         self.graphicsPath = None
@@ -50,7 +39,7 @@ class Background(Resource):
                 self.imageWidth = self.group.readInteger('Width')
                 self.imageHeight = self.group.readInteger('Height')
             except Exception as exc:
-                logException(exc) # TODO: simplify if we switch to twisted logger
+                logException(exc)  # TODO: simplify if we switch to twisted logger
                 raise
         self.isPlain = bool(self.group.readEntry('Plain'))
         if not self.isPlain:
@@ -62,7 +51,7 @@ class Background(Resource):
                     (graphName, self.desktopFileName))
 
     def pixmap(self, size):
-        """returns a background pixmap or None for isPlain"""
+        """return a background pixmap or None for isPlain"""
         self.__pmap = QBrush()
         if not self.isPlain:
             width = size.width()
@@ -89,7 +78,7 @@ class Background(Resource):
         return QBrush(self.pixmap(size))
 
     def setPalette(self, onto):
-        """sets a background palette for widget onto"""
+        """set a background palette for widget onto"""
         palette = QPalette()
         mybrush = self.brush(onto.size())
         palette.setBrush(QPalette.Window, mybrush)
